from pytezos.michelson.instructions.adt import CarInstruction
from pytezos.michelson.instructions.adt import CdrInstruction
from pytezos.michelson.instructions.adt import GetnInstruction
from pytezos.michelson.instructions.adt import LeftInstruction
from pytezos.michelson.instructions.adt import PairInstruction
from pytezos.michelson.instructions.adt import RightInstruction
from pytezos.michelson.instructions.adt import UnpairInstruction
from pytezos.michelson.instructions.adt import UpdatenInstruction
from pytezos.michelson.instructions.arithmetic import AbsInstruction
from pytezos.michelson.instructions.arithmetic import AddInstruction
from pytezos.michelson.instructions.arithmetic import EdivInstruction
from pytezos.michelson.instructions.arithmetic import IntInstruction
from pytezos.michelson.instructions.arithmetic import IsNatInstruction
from pytezos.michelson.instructions.arithmetic import LslInstruction
from pytezos.michelson.instructions.arithmetic import LsrInstruction
from pytezos.michelson.instructions.arithmetic import MulInstruction
from pytezos.michelson.instructions.arithmetic import NegInstruction
from pytezos.michelson.instructions.arithmetic import SubInstruction
from pytezos.michelson.instructions.boolean import AndInstruction
from pytezos.michelson.instructions.boolean import NotInstruction
from pytezos.michelson.instructions.boolean import OrInstruction
from pytezos.michelson.instructions.boolean import XorInstruction
from pytezos.michelson.instructions.compare import CompareInstruction
from pytezos.michelson.instructions.compare import EqInstruction
from pytezos.michelson.instructions.compare import GeInstruction
from pytezos.michelson.instructions.compare import GtInstruction
from pytezos.michelson.instructions.compare import LeInstruction
from pytezos.michelson.instructions.compare import LtInstruction
from pytezos.michelson.instructions.compare import NeqInstruction
from pytezos.michelson.instructions.control import ApplyInstruction
from pytezos.michelson.instructions.control import DipInstruction
from pytezos.michelson.instructions.control import DipnInstruction
from pytezos.michelson.instructions.control import ExecInstruction
from pytezos.michelson.instructions.control import FailwithInstruction
from pytezos.michelson.instructions.control import IfConsInstruction
from pytezos.michelson.instructions.control import IfInstruction
from pytezos.michelson.instructions.control import IfLeftInstruction
from pytezos.michelson.instructions.control import IfNoneInstruction
from pytezos.michelson.instructions.control import IterInstruction
from pytezos.michelson.instructions.control import LambdaInstruction
from pytezos.michelson.instructions.control import LoopInstruction
from pytezos.michelson.instructions.control import LoopLeftInstruction
from pytezos.michelson.instructions.control import MapInstruction
from pytezos.michelson.instructions.control import PushInstruction
from pytezos.michelson.instructions.crypto import Blake2bInstruction
from pytezos.michelson.instructions.crypto import CheckSignatureInstruction
from pytezos.michelson.instructions.crypto import HashKeyInstruction
from pytezos.michelson.instructions.crypto import KeccakInstruction
from pytezos.michelson.instructions.crypto import PairingCheckInstruction
from pytezos.michelson.instructions.crypto import SaplingEmptyStateInstruction
from pytezos.michelson.instructions.crypto import SaplingVerifyUpdateInstruction
from pytezos.michelson.instructions.crypto import Sha3Instruction
from pytezos.michelson.instructions.crypto import Sha256Instruction
from pytezos.michelson.instructions.crypto import Sha512Instruction
from pytezos.michelson.instructions.generic import ConcatInstruction
from pytezos.michelson.instructions.generic import NeverInstruction
from pytezos.michelson.instructions.generic import PackInstruction
from pytezos.michelson.instructions.generic import SizeInstruction
from pytezos.michelson.instructions.generic import SliceInstruction
from pytezos.michelson.instructions.generic import UnitInstruction
from pytezos.michelson.instructions.generic import UnpackInstruction
from pytezos.michelson.instructions.jupyter import BeginInstruction
from pytezos.michelson.instructions.jupyter import BigMapDiffInstruction
from pytezos.michelson.instructions.jupyter import CommitInstruction
from pytezos.michelson.instructions.jupyter import DebugInstruction
from pytezos.michelson.instructions.jupyter import DropAllInstruction
from pytezos.michelson.instructions.jupyter import DumpAllInstruction
from pytezos.michelson.instructions.jupyter import DumpInstruction
from pytezos.michelson.instructions.jupyter import PatchInstruction
from pytezos.michelson.instructions.jupyter import PatchValueInstruction
from pytezos.michelson.instructions.jupyter import PrintInstruction
from pytezos.michelson.instructions.jupyter import ResetInstruction
from pytezos.michelson.instructions.jupyter import ResetValueInstruction
from pytezos.michelson.instructions.jupyter import RunInstruction
from pytezos.michelson.instructions.stack import DigInstruction
from pytezos.michelson.instructions.stack import DropInstruction
from pytezos.michelson.instructions.stack import DropnInstruction
from pytezos.michelson.instructions.stack import DugInstruction
from pytezos.michelson.instructions.stack import DupInstruction
from pytezos.michelson.instructions.stack import DupnInstruction
from pytezos.michelson.instructions.stack import PushInstruction
from pytezos.michelson.instructions.stack import RenameInstruction
from pytezos.michelson.instructions.stack import SwapInstruction
from pytezos.michelson.instructions.struct import ConsInstruction
from pytezos.michelson.instructions.struct import EmptyBigMapInstruction
from pytezos.michelson.instructions.struct import EmptyMapInstruction
from pytezos.michelson.instructions.struct import EmptySetInstruction
from pytezos.michelson.instructions.struct import GetAndUpdateInstruction
from pytezos.michelson.instructions.struct import GetInstruction
from pytezos.michelson.instructions.struct import MemInstruction
from pytezos.michelson.instructions.struct import NilInstruction
from pytezos.michelson.instructions.struct import NoneInstruction
from pytezos.michelson.instructions.struct import SomeInstruction
from pytezos.michelson.instructions.struct import UpdateInstruction
from pytezos.michelson.instructions.tezos import AddressInstruction
from pytezos.michelson.instructions.tezos import AmountInstruction
from pytezos.michelson.instructions.tezos import BalanceInstruction
from pytezos.michelson.instructions.tezos import ChainIdInstruction
from pytezos.michelson.instructions.tezos import ContractInstruction
from pytezos.michelson.instructions.tezos import CreateContractInstruction
from pytezos.michelson.instructions.tezos import ImplicitAccountInstruction
from pytezos.michelson.instructions.tezos import MinBlockTimeInstruction
from pytezos.michelson.instructions.tezos import NowInstruction
from pytezos.michelson.instructions.tezos import SelfAddressInstruction
from pytezos.michelson.instructions.tezos import SelfInstruction
from pytezos.michelson.instructions.tezos import SenderInstruction
from pytezos.michelson.instructions.tezos import SetDelegateInstruction
from pytezos.michelson.instructions.tezos import SourceInstruction
from pytezos.michelson.instructions.tezos import TotalVotingPowerInstruction
from pytezos.michelson.instructions.tezos import TransferTokensInstruction
from pytezos.michelson.instructions.tezos import VotingPowerInstruction
from pytezos.michelson.instructions.ticket import JoinTicketsInstruction
from pytezos.michelson.instructions.ticket import ReadTicketInstruction
from pytezos.michelson.instructions.ticket import SplitTicketInstruction
from pytezos.michelson.instructions.ticket import TicketInstruction
from pytezos.michelson.instructions.tzt import BigMapInstruction
from pytezos.michelson.instructions.tzt import StackEltInstruction
